==> pthreads-1_50_3.tar.gz: pthreads/pthreads/mutex.c <==
/* ==== mutex.c ============================================================
 * Copyright (c) 1993, 1994 by Chris Provenzano, proven@mit.edu
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by Chris Provenzano.
 * 4. The name of Chris Provenzano may not be used to endorse or promote 
 *	  products derived from this software without specific prior written
 *	  permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CHRIS PROVENZANO ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL CHRIS PROVENZANO BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 *
 * Description : Mutex functions.
 *
 *  1.00 93/07/19 proven
 *      -Started coding this file.
 */

#ifndef lint
static const char rcsid[] = "mutex.c,v 1.50 1994/08/08 03:45:11 proven Exp";
#endif

#include <pthread.h>
#include <errno.h>

/*
 * Basic mutex functionality

 * This is the basic lock order
 * queue
 * pthread
 * global
 *
 * semaphore functionality is defined in machdep.h
 */

/* ==========================================================================
 * pthread_mutex_init()
 *
 * In this implementation I don't need to allocate memory.
 * ENOMEM, EAGAIN should never be returned. Arch that have
 * weird constraints may need special coding.
 */
int pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *mutex_attr)
{
	semaphore *lock;

	/* Only check if attr specifies some mutex type other than fast */
	if ((mutex_attr) && (mutex_attr->m_type != MUTEX_TYPE_FAST)) {
		if (mutex_attr->m_type >= MUTEX_TYPE_MAX) {
			return(EINVAL);
		}
		if (mutex->m_flags & MUTEX_FLAGS_INITED) {	
			return(EBUSY);
		}
		mutex->m_type = mutex_attr->m_type;
	} else {
		mutex->m_type = MUTEX_TYPE_FAST;
	}
	/* Set all other paramaters */
	pthread_queue_init(&mutex->m_queue);
	mutex->m_flags 	|= MUTEX_FLAGS_INITED;
	mutex->m_owner	= NULL;

	lock = &(mutex->m_lock);
	SEMAPHORE_RESET(lock);
	return(OK);
}

/* ==========================================================================
 * pthread_mutex_destroy()
 */
int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
	semaphore *lock;

	/* Only check if mutex is of type other than fast */
	switch(mutex->m_type) {
	case MUTEX_TYPE_FAST:
		break;
	case MUTEX_TYPE_STATIC_FAST:
	default:
		return(EINVAL);
		break;
	}

	/* Cleanup mutex, others might want to use it. */
	pthread_queue_init(&mutex->m_queue);
	mutex->m_flags 	|= MUTEX_FLAGS_INITED;
	mutex->m_owner	= NULL;
	mutex->m_flags	= 0;

	lock = &(mutex->m_lock);
	SEMAPHORE_RESET(lock);
	return(OK);
}

/* ==========================================================================
 * pthread_mutex_trylock()
 */
int pthread_mutex_trylock(pthread_mutex_t *mutex)
{
	semaphore *lock;
	int rval;

	lock = &(mutex->m_lock);
	while (SEMAPHORE_TEST_AND_SET(lock)) {
		pthread_yield();
	}
	
	switch (mutex->m_type) {
	/*
	 * Fast mutexes do not check for any error conditions.
     */
	case MUTEX_TYPE_FAST: 
	case MUTEX_TYPE_STATIC_FAST:
		if (!mutex->m_owner) {
			mutex->m_owner = pthread_run;
			rval = OK;
		} else {
			rval = EBUSY;
		}
		break;
	default:
		rval = EINVAL;
		break;
	}
	SEMAPHORE_RESET(lock);
	return(rval);
}

/* ==========================================================================
 * pthread_mutex_lock()
 */
int pthread_mutex_lock(pthread_mutex_t *mutex)
{
	semaphore *lock, *plock;
	int rval;

	lock = &(mutex->m_lock);
	while (SEMAPHORE_TEST_AND_SET(lock)) {
		pthread_yield();
	}
	
	switch (mutex->m_type) {
	/*
	 * Fast mutexes do not check for any error conditions.
     */
	case MUTEX_TYPE_FAST: 
	case MUTEX_TYPE_STATIC_FAST:
		if (mutex->m_owner) {
			plock = &(pthread_run->lock);
			while (SEMAPHORE_TEST_AND_SET(plock)) {
				 pthread_yield();
			}
			pthread_queue_enq(&mutex->m_queue, pthread_run);
			SEMAPHORE_RESET(lock);

			/* Reschedule will unlock pthread_run */
			reschedule(PS_MUTEX_WAIT);
			return(OK);
		}
		mutex->m_owner = pthread_run;
		rval = OK;
		break;
	default:
		rval = EINVAL;
		break;
	}
	SEMAPHORE_RESET(lock);
	return(rval);
}

/* ==========================================================================
 * pthread_mutex_unlock()
 */
int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
	struct pthread *pthread;
	semaphore *lock, *plock;
	int rval;

	lock = &(mutex->m_lock);
	while (SEMAPHORE_TEST_AND_SET(lock)) {
		pthread_yield();
	}

	switch (mutex->m_type) {
    /*
     * Fast mutexes do not check for any error conditions.
     */
    case MUTEX_TYPE_FAST:
    case MUTEX_TYPE_STATIC_FAST:
		if (pthread = pthread_queue_get(&mutex->m_queue)) {
			plock = &(pthread->lock);
			while (SEMAPHORE_TEST_AND_SET(plock)) {
				 pthread_yield();
			}
			mutex->m_owner = pthread;

			/* Reset pthread state */
			pthread_queue_deq(&mutex->m_queue);
			pthread->state = PS_RUNNING;
			SEMAPHORE_RESET(plock);
		} else {
			mutex->m_owner = NULL;
		}
		rval = OK;
		break;
	default:
		rval = EINVAL;
		break;
	}
	SEMAPHORE_RESET(lock);
	return(rval);
}

==> pthreads-1_51_0.tar.gz: pthreads/pthreads/mutex.c <==
/* ==== mutex.c ============================================================
 * Copyright (c) 1993, 1994 by Chris Provenzano, proven@mit.edu
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by Chris Provenzano.
 * 4. The name of Chris Provenzano may not be used to endorse or promote 
 *	  products derived from this software without specific prior written
 *	  permission.
 *
 * THIS SOFTWARE IS PROVIDED BY CHRIS PROVENZANO ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL CHRIS PROVENZANO BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE.
 *
 * Description : Mutex functions.
 *
 *  1.00 93/07/19 proven
 *      -Started coding this file.
 */

#ifndef lint
static const char rcsid[] = "mutex.c,v 1.50.2.1 1994/11/02 02:11:28 proven Exp";
#endif

#include <pthread.h>
#include <errno.h>

/*
 * Basic mutex functionality

 * This is the basic lock order
 * queue
 * pthread
 * global
 *
 * semaphore functionality is defined in machdep.h
 */

/* ==========================================================================
 * pthread_mutex_init()
 *
 * In this implementation I don't need to allocate memory.
 * ENOMEM, EAGAIN should never be returned. Arch that have
 * weird constraints may need special coding.
 */
int pthread_mutex_init(pthread_mutex_t *mutex, const pthread_mutexattr_t *mutex_attr)
{
	enum pthread_mutextype type;
	semaphore *lock;

	/* Only check if attr specifies some mutex type other than fast */
	if ((mutex_attr) && (mutex_attr->m_type != MUTEX_TYPE_FAST)) {
		if (mutex_attr->m_type >= MUTEX_TYPE_MAX) {
			return(EINVAL);
		}
		type = mutex_attr->m_type;
		if (mutex->m_flags & MUTEX_FLAGS_INITED) {	
			return(EBUSY);
		}
		mutex->m_type = mutex_attr->m_type;
	} else {
		type =  MUTEX_TYPE_FAST;
	}

	lock = &(mutex->m_lock);
	while (SEMAPHORE_TEST_AND_SET(lock)) {
		pthread_yield();
	}
	
	switch(type) {
	case MUTEX_TYPE_STATIC_FAST:
		SEMAPHORE_RESET(lock);
		return(EINVAL);
		break;
	case MUTEX_TYPE_FAST:
		break;
	case MUTEX_TYPE_DEBUG:
		if (mutex->m_flags & MUTEX_FLAGS_INITED) {
			SEMAPHORE_RESET(lock);
			return(EBUSY);
		}
		break;
	default:
		SEMAPHORE_RESET(lock);
		return(EINVAL);
		break;
	}

	/* Set all other paramaters */
	pthread_queue_init(&mutex->m_queue);
	mutex->m_flags 	|= MUTEX_FLAGS_INITED;
	mutex->m_owner	= NULL;
	mutex->m_type 	= type;

	SEMAPHORE_RESET(lock);
	return(OK);
}

/* ==========================================================================
 * pthread_mutex_destroy()
 */
int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
	semaphore *lock;

	lock = &(mutex->m_lock);
	while (SEMAPHORE_TEST_AND_SET(lock)) {
		pthread_yield();
	}
	
	/* Only check if mutex is of type other than fast */
	switch(mutex->m_type) {
	case MUTEX_TYPE_STATIC_FAST:
		SEMAPHORE_RESET(lock);
		return(EINVAL);
		break;
	case MUTEX_TYPE_FAST:
		break;
	case MUTEX_TYPE_DEBUG:
		if (!(mutex->m_flags & MUTEX_FLAGS_INITED)) {
			SEMAPHORE_RESET(lock);
			return(EINVAL);
		}
		if (mutex->m_owner) {
			SEMAPHORE_RESET(lock);
			return(EBUSY);
		}
		break;
	default:
		SEMAPHORE_RESET(lock);
		return(EINVAL);
		break;
	}

	/* Cleanup mutex, others might want to use it. */
	pthread_queue_init(&mutex->m_queue);
	mutex->m_owner	= NULL;
	mutex->m_flags	= 0;

	SEMAPHORE_RESET(lock);
	return(OK);
}

/* ==========================================================================
 * pthread_mutex_trylock()
 */
int pthread_mutex_trylock(pthread_mutex_t *mutex)
{
	semaphore *lock;
	int rval;

	lock = &(mutex->m_lock);
	while (SEMAPHORE_TEST_AND_SET(lock)) {
		pthread_yield();
	}
	
	switch (mutex->m_type) {
	/*
	 * Fast mutexes do not check for any error conditions.
     */
	case MUTEX_TYPE_FAST: 
	case MUTEX_TYPE_STATIC_FAST:
		if (!mutex->m_owner) {
			mutex->m_owner = pthread_run;
			rval = OK;
		} else {
			rval = EBUSY;
		}
		break;
	case MUTEX_TYPE_DEBUG:
		if (mutex->m_flags & MUTEX_FLAGS_INITED) {
			if (!mutex->m_owner) {
				mutex->m_owner = pthread_run;
				rval = OK;
			} else {
				rval = EBUSY;
			}
		} else {
			rval = EINVAL;
		}
		break;
	default:
		rval = EINVAL;
		break;
	}
	SEMAPHORE_RESET(lock);
	return(rval);
}

/* ==========================================================================
 * pthread_mutex_lock()
 */
int pthread_mutex_lock(pthread_mutex_t *mutex)
{
	semaphore *lock, *plock;
	int rval;

	lock = &(mutex->m_lock);
	while (SEMAPHORE_TEST_AND_SET(lock)) {
		pthread_yield();
	}
	
	switch (mutex->m_type) {
	/*
	 * Fast mutexes do not check for any error conditions.
     */
	case MUTEX_TYPE_FAST: 
	case MUTEX_TYPE_STATIC_FAST:
		if (mutex->m_owner) {
			plock = &(pthread_run->lock);
			while (SEMAPHORE_TEST_AND_SET(plock)) {
				 pthread_yield();
			}
			pthread_queue_enq(&mutex->m_queue, pthread_run);
			SEMAPHORE_RESET(lock);

			/* Reschedule will unlock pthread_run */
			reschedule(PS_MUTEX_WAIT);
			return(OK);
		}
		mutex->m_owner = pthread_run;
		rval = OK;
		break;
	case MUTEX_TYPE_DEBUG:
		if (mutex->m_flags & MUTEX_FLAGS_INITED) {
			if (mutex->m_owner) {
				if (mutex->m_owner != pthread_run) {
					plock = &(pthread_run->lock);
					while (SEMAPHORE_TEST_AND_SET(plock)) {
						 pthread_yield();
					}
					pthread_queue_enq(&mutex->m_queue, pthread_run);
					SEMAPHORE_RESET(lock);
		
					/* Reschedule will unlock pthread_run */
					reschedule(PS_MUTEX_WAIT);
			
					if (mutex->m_owner != pthread_run) {
						PANIC();
					}
					return(OK);
				}
				rval = EDEADLK;
				break;
			}
			mutex->m_owner = pthread_run;
			rval = OK;
			break;
		}
		rval = EINVAL;
		break;
	default:
		rval = EINVAL;
		break;
	}
	SEMAPHORE_RESET(lock);
	return(rval);
}

/* ==========================================================================
 * pthread_mutex_unlock()
 */
int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
	struct pthread *pthread;
	semaphore *lock, *plock;
	int rval;

	lock = &(mutex->m_lock);
	while (SEMAPHORE_TEST_AND_SET(lock)) {
		pthread_yield();
	}

	switch (mutex->m_type) {
    /*
     * Fast mutexes do not check for any error conditions.
     */
    case MUTEX_TYPE_FAST:
    case MUTEX_TYPE_STATIC_FAST:
		if (pthread = pthread_queue_get(&mutex->m_queue)) {
			plock = &(pthread->lock);
			while (SEMAPHORE_TEST_AND_SET(plock)) {
				 pthread_yield();
			}
			mutex->m_owner = pthread;

			/* Reset pthread state */
			pthread_queue_deq(&mutex->m_queue);
			pthread->state = PS_RUNNING;
			SEMAPHORE_RESET(plock);
		} else {
			mutex->m_owner = NULL;
		}
		rval = OK;
		break;
	case MUTEX_TYPE_DEBUG:
		if (mutex->m_owner == pthread_run) {
			if (pthread = pthread_queue_get(&mutex->m_queue)) {
				plock = &(pthread->lock);
				while (SEMAPHORE_TEST_AND_SET(plock)) {
					 pthread_yield();
				}
				mutex->m_owner = pthread;

				/* Reset pthread state */
				pthread_queue_deq(&mutex->m_queue);
				pthread->state = PS_RUNNING;
				SEMAPHORE_RESET(plock);
			} else {
				mutex->m_owner = NULL;
			}
			rval = OK;
		} else {
			rval = EPERM;
		}
		break;
	default:
		rval = EINVAL;
		break;
	}
	SEMAPHORE_RESET(lock);
	return(rval);
}
